/******************************************************************************
Asservissement de verticalité et de mouvement d'un gyropode.
Ce programme est destiné à fonctionner avec Geeros Wifi, Lynx ou RPi, 
disponibles à l'adresse suivante:
http://boutique.3sigma.fr/12-robots

ATTENTION: ce programme fonctionne uniquement avec les robots livrés
après le 1er octobre 2016

Auteur: 3Sigma
Version 3.0 - 10/10/2016
*******************************************************************************/


// Inclusion d'une bibliothèque permettant l'exécution à cadence fixe
// d'une partie du programme. Télécharger à l'adresse http://www.3sigma.fr/telechargements/FlexiTimer2.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
// Pour plus de détails, voir les pages (en anglais): 
// http://www.arduino.cc/playground/Main/FlexiTimer2
// https://github.com/wimleers/flexitimer2
#include <FlexiTimer2.h>

// Inclusion d'une bibliothèque permettant de lire et d'écrire plus rapidement sur les entrées-sorties digitales.
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/digitalWriteFast.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <digitalWriteFast.h> 

#include "MPU9250.h"

// Inclusion d'une bibliothèque permettant une commande synchronisée avec l'"interruption timer"
// de l'asservissement (voir plus loin) du servomoteur supportant la caméra
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/SoftwareServo.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <SoftwareServo.h> 

// Définitions et déclarations pour le codeur incrémental du moteur droit
#define codeurDroitInterruptionA 0
#define codeurDroitPinA 2
#define codeurDroitPinB 8
#define Nmoy 10
volatile long ticksCodeurDroit = 0;
static int indiceTicksCodeurDroit = 0;
static int ticksCodeurDroitTab[Nmoy];
static int codeurDroitDeltaPos;

// Définitions et déclarations pour le codeur incrémental du moteur gauche
#define codeurGaucheInterruptionA 1
#define codeurGauchePinA 3
#define codeurGauchePinB 9
volatile long ticksCodeurGauche = 0;
static int indiceTicksCodeurGauche = 0;
static int ticksCodeurGaucheTab[Nmoy];
static int codeurGaucheDeltaPos;

// Définitions et déclarations pour le moteur à courant continu.
#define directionMoteurDroit  4
#define pwmMoteurDroit  5
#define directionMoteurGauche  7
#define pwmMoteurGauche  6

// Déclarations pour le servomoteur
SoftwareServo servocam;

// Ce programme reçoit par liaison série des données issues d'un programme qui est exécuté sur
// l'ordinateur connecté à la carte. Si le programme ne reçoit rien pendant un temps donné,
// défini ci-dessous, un traitement particulier sera réalisé.
// Timeout de réception de données en s
#define TIMEOUT 2
unsigned long timeLastReceived = 0;
static int timedOut = 0;

// Ce programme envoie par liaison série des données à l'ordinateur connecté à la carte.
// Cadence d'envoi des données en ms
#define TSDATA 100
unsigned long tempsDernierEnvoi = 0;
unsigned long tempsCourant = 0;

// Certaines parties du programme sont exécutées à cadence fixe grâce à la bibliothèque FlexiTimer2.
// Cadence d'échantillonnage en ms
#define CADENCE_MS 10
volatile double dt = CADENCE_MS/1000.;
volatile double temps = -CADENCE_MS/1000.;

// Le gyropode est asservi grâce à plusieurs régulateurs de type PI
// On déclare ci-dessous les variables et paramètres nécessaires à l'asservissement et au régulateur
static double R = 0.045; // Rayon d'une roue
static double W = 0.14; // Largeur du gyropode
static double umax = 6.; // valeur max de la tension de commande du moteur
static double umin = -6.; // valeur min (ou max en négatif) de la tension de commande du moteur
static double vxmes = 0.; // vitesse longitudinale mesurée
static double omegames = 0.; // vitesse de rotation par rapport à la verticale mesurée
static double thetames = 0.; // angle par rapport à la verticale mesuré
static double xidotmes = 0.; // vitesse de rotation mesurée
static double Kpvx; // gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kpvx1 = -29.7; // sous-gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kpvx2 = 0.15; // sous-gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kivx = -30.4; // gain intégral pour l'asservissement de vitesse longitudinale
static double Kpomega = -3.03; // gain proportionnel pour l'asservissement de verticalité
static double Kiomega = -36.6; // gain intégral pour l'asservissement de verticalité
static double Kpxidot = 0.; // gain proportionnel pour l'asservissement de rotation
static double Kixidot = 2.14; // gain intégral pour l'asservissement de rotation
static double commande_vx = 0.; // commande pour l'asservissement de vitesse longitudinale
static double commande_omega = 0.; // commande pour l'asservissement de verticalité
static double commande_xidot = 0.; // commande pour l'asservissement de rotation
static double P_vx = 0.; // action proportionnelle pour l'asservissement de vitesse longitudinale
static double I_vx = 0.; // action intégrale pour l'asservissement de vitesse longitudinale
static double P_omega = 0.; // action proportionnelle pour l'asservissement de verticalité
static double I_omega = 0.; // action intégrale pour l'asservissement de verticalité
static double P_xidot = 0.; // action proportionnelle pour l'asservissement de rotation
static double I_xidot = 0.; // action intégrale pour l'asservissement de rotation
static double thetaest = 0.; // angle d'inclinaison estimé par le filtre complémentaire
static double tau = 1.; // paramètre du filtre complémentaire

// Déclarations pour la commande des moteurs
static double omegaDroit = 0.;
static double omegaGauche = 0.;
static double commandeDroit = 0.;
static double commandeGauche = 0.;

// Déclarations pour les consignes de mouvement
static double vxref = 0.;
static double xidotref = 0.;
static double omegaref = 0.;
static double thetaref = 0.;


// Déclarations pour les accélérations, angles et vitesses de rotation
static int16_t axint;
static int16_t gyroint;
static double ax;
static double omega;

// Déclarations pour la tension batterie
static double tensionBatterie;

// Déclarations pour la gestion des modes "asservissement actif" et "chute"
static int started = false;
static int juststarted = true;
static int signe_ax = 0;

// Déclaration des variables pour la réception des données sur la ligne série
#define NOMBRE_DE_CHAMPS 6 // nombre de champs de données reçues sur la liaison série
int champIndex = 0; // champ courant reçu
int valeursRecues[NOMBRE_DE_CHAMPS]; // tableau contenant les champs de données reçues
// Variables utilisées pour les données reçues
static double x1 = 0.;
static double x2 = 0.;
static int servoref = 45;
static double Kp2 = 1.;
static double Ki2 = 1.;
// Variables pour la vérification de cohérence des données
static int crc = 0;
static int crc_check = 0;

// Déclaration de l'objet "capteur accélérométrique et gyroscopique"
MPU9250 accelgyro;

// Initialisations
void setup(void) {
  
  // Liaison série.
  // ATTENTION: ne pas modifier la vitesse de transmission de la liaison série,
  Serial.begin(57600);
  Serial.flush();
  
  Wire.begin();
  // TWBR = 12;  // Vitesse I2C de 400 kbit/sec

  // Lecture du registre WHO_AM_I, ce qui est un bon test de communication
  byte c = accelgyro.readByte(MPU9250_ADDRESS, WHO_AM_I_MPU9250);
  Serial.print("MPU9250 "); Serial.print("I AM "); Serial.print(c, HEX);
  Serial.print(" Je devrais etre "); Serial.println(0x71, HEX);


  if (c == 0x71) {
    Serial.println("MPU9250 est en ligne...");

    // Auto-test
    accelgyro.MPU9250SelfTest(accelgyro.SelfTest);
    Serial.print("Auto test axe X : trim acceleration dans : ");
    Serial.print(accelgyro.SelfTest[0],1); Serial.println("% des valeurs d'usine");
    Serial.print("Auto test axe Y : trim acceleration dans : ");
    Serial.print(accelgyro.SelfTest[1],1); Serial.println("% des valeurs d'usine");
    Serial.print("Auto test axe Z : trim acceleration dans : ");
    Serial.print(accelgyro.SelfTest[2],1); Serial.println("% des valeurs d'usine");
    Serial.print("Auto test axe X : trim gyro dans : ");
    Serial.print(accelgyro.SelfTest[3],1); Serial.println("% des valeurs d'usine");
    Serial.print("Auto test axe Y : trim gyro dans : ");
    Serial.print(accelgyro.SelfTest[4],1); Serial.println("% des valeurs d'usine");
    Serial.print("Auto test axe Z : trim gyro dans : ");
    Serial.print(accelgyro.SelfTest[5],1); Serial.println("% des valeurs d'usine");

    // Calibration gyro et accéléro, chargement des biais dans les registres de biais
    //accelgyro.calibrateMPU9250(accelgyro.gyroBias, accelgyro.accelBias);


    accelgyro.initMPU9250();
    Serial.println("MPU9250 initialise pour mode actif....");

    // Resolutions
    accelgyro.getAres();
    accelgyro.getGres();

    Serial.print("Resolution accel: ");
    Serial.println(accelgyro.aRes, 6);
    Serial.print("Resolution gyro: ");
    Serial.println(accelgyro.gRes, 6);


  } // if (c == 0x71)
  else {
    Serial.print("Impossible de se connecter au MPU9250: 0x");
    Serial.println(c, HEX);
    while(1) ; // Boucle infinie si la communication a échoué
  }

  
  // Codeur incrémental moteur droit
  pinMode(codeurDroitPinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurDroitPinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurDroitPinA (définie à la fin du programme)
  attachInterrupt(codeurDroitInterruptionA, GestionInterruptionCodeurDroitPinA, CHANGE);

  // Codeur incrémental moteur gauche
  pinMode(codeurGauchePinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurGauchePinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurGauchePinA (définie à la fin du programme)
  attachInterrupt(codeurGaucheInterruptionA, GestionInterruptionCodeurGauchePinA, CHANGE);
 
  // Moteur à courant continu droit
  pinMode(directionMoteurDroit, OUTPUT);
  pinMode(pwmMoteurDroit, OUTPUT);

  // Moteur à courant continu gauche
  pinMode(directionMoteurGauche, OUTPUT);
  pinMode(pwmMoteurGauche, OUTPUT);
  
  // Servomoteur
  servocam.attach(10);

  // Compteur d'impulsions des encodeurs
  ticksCodeurDroit = 0;
  ticksCodeurGauche = 0;
        
  // On démarre seulement quand le gyropode dépasse la verticale
  // Pour que le gyropode démarre tout seul quand il est sur l'avant
  // et si on le redresse quand il est sur l'arrière
  signe_ax = -1;
  
  
  /* Le programme principal est constitué:
     - de la traditionnelle boucle "loop" des programmes Arduino, dans laquelle on lit 
       en permanence la liaison série pour récupérer les nouvelles consignes de mouvement
     - de la fonction "isrt", dont l'exécution est définie à cadence fixe par les deux instructions
       ci-dessous. La bonne méthode pour exécuter une fonction à cadence fixe, c'est de l'exécuter sur interruption
       (la boucle "loop" est alors interrompue régulièrement, à la cadence voulue, par un timer).
       Une mauvaise méthode serait d'utiliser la fonction Arduino "delay". En effet, la fonction "delay" définit
       un délai entre la fin des instructions qui la précèdent et le début des instructions qui la suivent, mais si
       on ne connait pas le temps d'exécution de ces instructions (ou si celui-ci est inconnu), on n'obtiendra
       jamais une exécution à cadence fixe.
       Il est nécessaire d'exécuter la fonction "isrt" à cadence fixe car cette fonction:
       - doit renvoyer des données à cadence fixe sur la liaison série, pour monitoring
       - calcule l'asservissement de verticalité et de mouvement, conçu pour fonctionner à une cadence bien spécifiée
     
  */
  FlexiTimer2::set(CADENCE_MS, 1/1000., isrt); // résolution timer = 1 ms
  FlexiTimer2::start();

}


// Boucle principale
void loop() {

  // Tant que l'asservissement de verticalité n'est pas activé
  while (started==false) {
    
    // Geeros vient de chuter
    if (juststarted==true) {
      
      // On réinitialise l'asservissement et on entre dans la boucle d'attente de redressement
      //Serial.print("\r\r\rBoucle d'attente\r\r\r");
      juststarted = false;
      
      // Réinitilisation de l'asservissement
      I_vx = 0.;
      I_omega = 0.;
      I_xidot = 0.;
      thetaest = 0.;

    }
    
    // Mesure de la pesanteur
    accelgyro.readAccelData(accelgyro.accelCount);
    axint = accelgyro.accelCount[1];
    ax = -((float)axint * accelgyro.aRes) * 9.81;

    // Si l'accélération change de signe, cela signifie que Geeros a été redressé pour le faire démarrer
    if (ax*signe_ax<0) {
      
      // Donc, on démarre
      started = true;
      //Serial.print("\r\r\rStarted\r\r\r");
      
      // Les consignes reçues sont initilisées à zéro
      x1 = 0.;
      x2 = 0.;
      
      // On réinitialise le comptage d'impulsions des encodeurs
      // On aurait pu désactiver les interruptions pendant cette phase
      // mais ce qu'on veut au final, c'est cette init
      ticksCodeurDroit = 0;
      ticksCodeurGauche = 0;
    
    }
    else {
      // On garde tout à zéro
      I_vx = 0.;
      I_omega = 0.;
      I_xidot = 0.;
      thetaest = 0.;
    }
  }


  // Réception des données sur la liaison série
  readData();

}

// Fonction excutée sur interruption
void isrt(){

  // Compteur de boucle et variable d'activation de l'asservissement
  int i;
  double en = 0.; 

  // Mesure de la vitesse des moteurs grâce aux codeurs incrémentaux
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurDroitPinA et GestionInterruptionCodeurDroitPinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurDroitTab[i] += ticksCodeurDroit;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurDroit = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurDroitDeltaPos = ticksCodeurDroitTab[indiceTicksCodeurDroit];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurDroitTab[indiceTicksCodeurDroit] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurDroit++;
  if (indiceTicksCodeurDroit==Nmoy) {
    indiceTicksCodeurDroit = 0;
  }
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurGauchePinA et GestionInterruptionCodeurGauchePinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurGaucheTab[i] += ticksCodeurGauche;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurGauche = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurGaucheDeltaPos = ticksCodeurGaucheTab[indiceTicksCodeurGauche];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurGaucheTab[indiceTicksCodeurGauche] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurGauche++;
  if (indiceTicksCodeurGauche==Nmoy) {
    indiceTicksCodeurGauche = 0;
  }

  // Calcul de la vitesse de rotation. C'est le nombre d'impulsions converti en radian, divisé par la période d'échantillonnage
  // On fait un calcul glissant sur Nmoy échantillons, d'où le Nmoy*dt
  // ATTENTION: le facteur 2 devant est là car avec 2 moteurs, la résolution est divisée par 2
  omegaDroit = -2*((2.*3.141592*((double)codeurDroitDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s
  omegaGauche = 2*((2.*3.141592*((double)codeurGaucheDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s

  
  // Vitesse de rotation mesurée par le gyroscope convertie en rad/s
  omega = -(float)gyroint * accelgyro.gRes * 3.141592 / 180.;

  // Acceleration longitudinale
  thetames = -((float)axint * accelgyro.aRes);
  
  // Estimation de l'angle à partir de l'accélération horizontale (filtre complémentaire)
  thetaest = (tau*dt*omega + dt*thetames + tau*thetaest)/(dt+tau);
  
  // Test pour savoir si le gyropode est toujours debout
  if (started == true) {
    
    // Si l'angle d'inclinaison n'est pas trop prononcé
    if (abs(thetames) < 1.) {
      en = 1.;
    }
    
    // Sinon, cela signifie que Geeros est tombé
    else {
      // On désactive l'asservissement
      en = 0.;
      started = false;
      juststarted = true;
      //Serial.print("\r\r\rChute\r\r\r");
      
      // Réinitialisation pour détecter un futur redressement
      if (thetames>0) {
        signe_ax = 1;
      }
      else {
        signe_ax = -1;
      }
    }
  }
  
  // Si Geeros n'a pas encore démarré, l'asservissement est désactivé
  else {
    en = 0.;
  }

  // Application de la consigne lue sur la liaison série
  // x1 et x2 sont définis dans la fonction readData (voir plus loin)
  vxref = x1 * en;
  xidotref = x2 * en;

  
  // Définition des entrées de la fonction d'asservissement
  vxmes = (omegaDroit + omegaGauche)*R/2 * en;
  omegames = omega * en;
  xidotmes = (omegaDroit - omegaGauche)*R/W * en;

  /******* Calcul du PI sur vx *******/
  // Kp2 et Ki2 sont définis dans la fonction readData (voir plus loin)
  Kpvx = (Kpvx1 + Kpvx2/R) * Kp2;
  
  // Terme proportionnel (la transformation de la commande par retour d'état en PI
  // conduit à une référence nulle, d'où le 0.*vxref)
  P_vx = Kpvx * (0.*vxref - vxmes);
  
  // Calcul de la commande avant saturation
  commande_vx = P_vx + I_vx;
    
  // Terme intégral (sera utilisé lors du pas d'échantillonnage suivant)
  I_vx = I_vx + Kivx * Ki2 * dt * (vxref - vxmes);

  /******* Fin Calcul du PI sur vx *******/

  /******* Calcul du PI sur omega *******/  
  // Terme proportionnel
  P_omega = Kpomega * (omegaref - omegames);
  
  // Terme "intégral" (c'est en fait un terme proportionnel sur theta)
  I_omega = Kiomega * (thetaref - thetaest);
  
  // Calcul de la commande avant saturation
  commande_omega = P_omega + I_omega;
  
  /******* Fin Calcul du PI sur omega *******/
  
  /******* Calcul du PI sur xidot *******/
  // Terme proportionnel
  P_xidot = Kpxidot * (xidotref - xidotmes);
  
  // Calcul de la commande avant saturation
  commande_xidot = P_xidot + I_xidot;
  
  // Terme intégral (sera utilisé lors du pas d'échantillonnage suivant)
  I_xidot = I_xidot + Kixidot * dt * (xidotref - xidotmes);

  /******* Fin Calcul du PI sur xidot *******/

  
  // Calcul des commandes des moteurs
  commandeDroit = (commande_vx + commande_omega + commande_xidot) * en;
  commandeGauche = (commande_vx + commande_omega - commande_xidot) * en;
  
  // Application des commandes aux moteurs
  tensionBatterie = 7.4;
  CommandeMoteurDroit(-commandeDroit, tensionBatterie);
  CommandeMoteurGauche(-commandeGauche, tensionBatterie);
  
  // Commande servomoteur
  // servoref est défini dans la fonction readData (voir plus loin)
  servocam.write(servoref);
  SoftwareServo::refresh();
  
  // Ecriture des données sur la liaison série pour monitoring
  ecritureData();
  
  // Incrémentation du temps courant
  temps += dt;

}


void ecritureData(void) {
  // Ecriture des données en sortie tous les TSDATA millisecondes
  
  // On envoie les données s'il s'est écoulé plus de TSDATA secondes
  // entre l'instant courant et le dernier envoi
  tempsCourant = millis();
  if (tempsCourant-tempsDernierEnvoi > TSDATA) {
    Serial.print(temps);
    
    Serial.print(",");
    Serial.print(commandeDroit);
    Serial.print(",");
    Serial.print(commandeGauche);
    Serial.print(",");
    Serial.print(omegaDroit);
    Serial.print(",");
    Serial.print(omegaGauche);
    Serial.print(",");
    Serial.print(omega);
    Serial.print(",");
    Serial.print(thetames);
    Serial.print(",");
    Serial.print(thetaest);
   
    Serial.print("\r");
    Serial.print("\n");
    
    tempsDernierEnvoi = tempsCourant;
  }  
}

void readData(void) {
  // Lecture des données sur la liaison série
  // On attend une série de NOMBRE_DE_CHAMPS valeurs entières séparées par des virgules
  
  // Initialisations
  champIndex = 0;
  for (int j=0; j < NOMBRE_DE_CHAMPS; j++) {
    valeursRecues[j] = 0;
  }

  while(true) {

    // Lecture de l'accélération et de la vitesse de rotation
    // Nécessaire de faire ce qui suit ici car cette boucle infinie
    // bloque tout le reste si aucune données n'est reçue
    accelgyro.readAccelData(accelgyro.accelCount);
    axint = accelgyro.accelCount[1];
    accelgyro.readGyroData(accelgyro.gyroCount);
    gyroint = accelgyro.gyroCount[0];
  
    // Si Geeros n'a pas démarré, ce n'est pas la peine d'aller plus loin
    if (started == false) {
      break;
    }
    
    if (Serial.available()>0) { // si une donnée est reçue
      // Réinitialisation du timeout (voir plus loin)
      timeLastReceived = millis();
      timedOut = 0;
      // Lecture de la donnée
      char ch = Serial.read();
      if (ch >= '0' && ch <= '9') { // caractère ascii entre 0 et 9 ?
        // Si oui, nous stockons la valeur
        valeursRecues[champIndex] = (valeursRecues[champIndex] * 10) + (ch - '0');
      }
      else if (ch == ',') { // séparateur virgule détecté, on passe au champ suivant
        if(champIndex < NOMBRE_DE_CHAMPS-1) {
          champIndex++; // incrémentation de l'index de champ
        }
      }
      else {
        // tout caractère autre qu'un chiffre ou une virgule (le retour chariot de fin de trame par exemple)
        // stoppe l'acquisition des champs
        break; // on sort de la boucle
      }
      
    }
    // Timeout
    // Si aucune donnée n'a été reçue pendant plus de TIMEOUT ms, on réinitialise toutes les
    // consignes par sécurité, sauf la commande du servomoteur
    else {      
      if ((millis()-timeLastReceived > TIMEOUT*1000) && (timedOut==0)) {
        x1 = 0.;
        x2 = 0.;
        Kp2 = 1.;
        Ki2 = 1.;
        //Serial.print("\r\r\rTimeout !!!\r\r\r");
        timedOut = 1;
      }
    }    
  }
  
  // Traitement des données reçues (on extrait chaque valeur de la trame complète)
  //Serial.println(champIndex);
  if (champIndex == NOMBRE_DE_CHAMPS-1) { // si le bon nombre de champs a été reçu
      
    // Récupération de la dernière valeur (somme de contrôle, envoyée par l'émetteur)
    crc = valeursRecues[NOMBRE_DE_CHAMPS-1];
    // Calcul de la somme de contrôle (somme de toutes les autres données) côté récepteur
    crc_check = 0;
    for (int j=0; j < NOMBRE_DE_CHAMPS-1; j++) {
      crc_check += valeursRecues[j];
    }
    
    // Si les sommes de contrôle sont identiques, on récupère les données
    if (crc == crc_check) {
      // Conversion des données d'entiers en flottants
      x1 = ((double)valeursRecues[0] - 50)/100.;
      x2 = 3*((double)valeursRecues[1] - 60)*3.141592/180.;
      servoref = valeursRecues[2];
      Kp2 = 1.+(3.-(double)valeursRecues[3])/20.;
      Ki2 = 1.+(3.-(double)valeursRecues[4])/20.;
    }
    // Si les sommes de contrôle sont différentes, on fait un traitement particulier
    else {
      // Par exemple, envoi d'un message, désactivé ici
      //Serial.print("\r\r\rErreur CRC !!!");
    }
  }
  else { // on n'a pas reçu le bon nombre de champs
    // On fait un traitement particulier, pas exemple, envoi d'un message (désactivé ici)
    //Serial.print("\r\r\rErreur reception !!!\r\r\r");
  }
  
}

void CommandeMoteurDroit(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeDroit(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }

  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurDroit, LOW);
    analogWrite(pwmMoteurDroit, tension_int);
  }

  if (tension_int<0) {
    digitalWrite(directionMoteurDroit, HIGH);
    analogWrite(pwmMoteurDroit, -tension_int);
  }
}


double redresse_commandeDroit(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}


void CommandeMoteurGauche(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeGauche(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }
  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurGauche, LOW);
    analogWrite(pwmMoteurGauche, tension_int);
  }
  if (tension_int<0) {
    digitalWrite(directionMoteurGauche, HIGH);
    analogWrite(pwmMoteurGauche, -tension_int);
  }
}


double redresse_commandeGauche(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}

 
void GestionInterruptionCodeurDroitPinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental droit
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurDroitPinA) == digitalReadFast2(codeurDroitPinB)) {
    ticksCodeurDroit++;
  }
  else {
    ticksCodeurDroit--;
  }
}


void GestionInterruptionCodeurGauchePinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental gauche
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurGauchePinA) == digitalReadFast2(codeurGauchePinB)) {
    ticksCodeurGauche++;
  }
  else {
    ticksCodeurGauche--;
  }
}




